/*
Module : ATLCPImplMT.h
Purpose: Implementation for a replacement class for the ATL class IConnectionPointImpl which performs 
         automatic marshaling of client interfaces acrosss COM apartments using the GIT (Global Interface
         Table). Put more simply, this class will allow you to fire events from your COM objects from
         a different thread than the one on which the object was initially created. The code was based
         on the MS Article "SAMPLE: ATLCPImplMT encapsulates ATL event firing across COM apartments" at
         http://support.microsoft.com/kb/280512. As mentioned in this article, because the code uses the 
         Global Interface Table, it requires Windows NT 4.0 Service Pack 3 or later, or Windows 95 (or later) 
         with DCOM version 1.1 or later to be installed. Based on this original code, I have performed an 
         extensive rewrite of the class using the ATL 8 "CComGITPtr" class. In addition it has been updated 
         based upon the ATL 8 "IConnectionPointImpl" implementation. It also uses ATL::CSimpleArray
         to hold the GIT'ed interfaces rather than the CComDynamicUnkArray class.
         
         Note that this class unlike the rest of the code in GPSCom is licensed as open-source / freeware, 
         meaning that you are free to use this class in your projects (subject to the copyright details 
         below), without being required to register GPSCom.
         
         To use the class in your ATL client code, you first need to perform the usual sequence of steps
         to support events in your ATL project. This will result in the creation of a proxy class which is
         derived from IConnectionPointImpl which contains a number of "Fire_{EventName}" methods which you 
         would then derive from in your normal ATL objects, thus allowing you to fire events back to your 
         clients. You should change the inheritance of this proxy class to be derived from IConnectionPointImplMT 
         instead of IConnectionPointImpl. Note that the template parameters for IConnectionPointImplMT are 
         slightly different than IConnectionPointImpl. The template parameters for IConnectionPointImplMT 
         are as follows:
         
         template <class T, const IID* piid, class TSinkInterface> class IConnectionPointImplMT.
         
         Also for each generated function (named Fire_{EventName}) within the proxy class, you will need to          
         rework the firing code similiar to the following: 

         HRESULT Fire_OnRMCSentence()
	       {
      		 CComCritSecLock<CComAutoCriticalSection> lock(m_CPMTCritSec);
	       
		       CComVariant varResult;
		       int nConnections = m_Clients.GetSize();
		       for (int nConnectionIndex = 0; nConnectionIndex < nConnections; nConnectionIndex++)
		       {
             CComPtr<IDispatch> sp;
             sp.Attach(GetInterfaceAt(nConnectionIndex));
			       if (sp)
			       {
			         Code to call sp->Invoke with correct parameters....
			         ......
			         ......
				       sp->Invoke(.....);
			        }
		        }
		        return .....;
	        }
         
	       
	       That's all you need to do. All other client remains the same. For a concrete example of using the 
	       class in client code, please see the GPSComCP.h file in the GPSCom download. Happy COM coding!
	       
Created: PJN / 21-06-2002
History: PJN / 20-07-2006 1. Updated code to clean compile on VC 2005
         PJN / 20-01-2007 1. Major rewrite of the code as documented above in the Purpose section
         PJN / 08-02-2007 1. Updated copyright details.
                          2. How you implement the Fire_ methods as documented above has been updated to refer
                          to the need to serialize access to the "m_Clients" member variable which is 
                          accessed via the GetInterfaceAt method. The CProxy_IGPSConnectionEvents class in the 
                          GPSCom project and the IConnectionPointImplMT::GetInterfaceAt method have been 
                          updated to address this issue. Thanks to Michael Ford for reporting this bug. 
                          3. The actual pointers which are stored on the internal m_Clients array are now
                          "CComGITPtr<TSinkInterface>" heap pointers rather than actual stack instances. By 
                          doing this, the code now has explicit control over when and how copy constructors and 
                          operator= methods are being called as we add a element to the array. This avoids extra
                          overhead as the interface pointer is being added to the GIT. It also addresses an issue
                          where the cookie value which was being returned by Advise was actually incorrect due to
                          the fact that a copy of the CComGITPtr was being added to the array. Thanks to
                          Michael Ford for reporting this issue.
         PJN / 21-03-2010 1. Updated copyright details.
                          2. Updated the instructions on how to code the Fire_* methods to avoid a COM reference 
                          counting problem which could have resulted in memory leaks. Thanks to Peter Blanz for
                          reporting this issue.

Copyright (c) 2002 - 2008 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


/////////////////////////// Macros / Defines //////////////////////////////////

#pragma once

#ifndef __CPIMPLMT_H__
#define __CPIMPLMT_H__


////////////////////////// Includes ///////////////////////////////////////////

#include <atlcom.h>


///////////////////////// Classes /////////////////////////////////////////////

template <class T, const IID* piid, class TSinkInterface>
class ATL_NO_VTABLE IConnectionPointImplMT : public _ICPLocator<piid>
{
	typedef CComEnum<IEnumConnections, &__uuidof(IEnumConnections), CONNECTDATA, _Copy<CONNECTDATA> > CComEnumConnections;

public:
//Methods
  ~IConnectionPointImplMT()
  {
    //Serialize access to the array
    CComCritSecLock<CComAutoCriticalSection> lock(m_CPMTCritSec);

    //Tidy up the client array
    int nClients = m_Clients.GetSize();
    for (int i=nClients-1; i>=0; i--)
    {
      CComGITPtr<TSinkInterface>* pClient = m_Clients[i];
      m_Clients.RemoveAt(i);
      delete pClient;
    }
  }

	STDMETHOD(_LocCPQueryInterface)(REFIID riid, void** ppvObject)
	{
#ifndef _ATL_OLEDB_CONFORMANCE_TESTS
		ATLASSERT(ppvObject != NULL);
#endif
		if (ppvObject == NULL)
			return E_POINTER;
		*ppvObject = NULL;

		if (InlineIsEqualGUID(riid, __uuidof(IConnectionPoint)) || InlineIsEqualUnknown(riid))
		{
			*ppvObject = this;
			AddRef();
#ifdef _ATL_DEBUG_INTERFACES
			_AtlDebugInterfacesModule.AddThunk(reinterpret_cast<IUnknown**>(ppvObject), _T("IConnectionPointImplMT"), riid);
#endif // _ATL_DEBUG_INTERFACES
			return S_OK;
		}
		else
			return E_NOINTERFACE;
	}

	STDMETHOD(GetConnectionInterface)(IID* piid2)
	{
		if (piid2 == NULL)
			return E_POINTER;
		*piid2 = *piid;
		return S_OK;
	}

	STDMETHOD(GetConnectionPointContainer)(IConnectionPointContainer** ppCPC)
	{
		T* pT = static_cast<T*>(this);
		
		//No need to check ppCPC for NULL since QI will do that for us
		return pT->QueryInterface(IID_IConnectionPointContainer, reinterpret_cast<void**>(ppCPC));
	}

	STDMETHOD(Advise)(IUnknown* pUnkSink, DWORD* pdwCookie)
  {
	  //Validate our parameter
	  if (pUnkSink == NULL || pdwCookie == NULL)
		  return E_POINTER;
		
		//Default the output parameters
		*pdwCookie = 0;

    //Check that the client implements the expected sink interface
    CComPtr<TSinkInterface> pClientSink;
    HRESULT hr = pUnkSink->QueryInterface(&pClientSink);  
	  if (SUCCEEDED(hr))
	  {
	    //Create the new GIT pointer on the heap (so that we can control when 
	    //and how copy constructors and operator= methods are being called as 
	    //we add a element to the array)
	    CComGITPtr<TSinkInterface>* pNewClient = NULL;
		  ATLTRY(pNewClient = new CComGITPtr<TSinkInterface>)
		  if (pNewClient == NULL)
		    return E_OUTOFMEMORY;
		    
	    hr = pNewClient->Attach(pClientSink);
	    if (hr == S_OK)
	    {
	      //And then get the corresponding cookie for the specified interface
	      *pdwCookie = pNewClient->GetCookie();
		    
	      //Serialize access to the array as we are about to add a new client
		    CComCritSecLock<CComAutoCriticalSection> lock(m_CPMTCritSec);
  		  
		    //Finally add to the array of clients
        if (!m_Clients.Add(pNewClient))
        {
          hr = E_FAIL;
          delete pNewClient;
        }
	    }
	    else
	      delete pNewClient;
	  }
	  else if (hr == E_NOINTERFACE)
		  hr = CONNECT_E_CANNOTCONNECT;

	  return hr;
  }

	STDMETHOD(Unadvise)(DWORD dwCookie)
  {
    //What will be the return value from this function (assume the best)
    HRESULT hr = S_OK;
    
    //Value used to control termination of the search loop below
    BOOL bFound = FALSE;
    
    //Serialize access to the array
    CComCritSecLock<CComAutoCriticalSection> lock(m_CPMTCritSec);

    //Remove the specified client from the array of clients, by matching up
    //the input cookie with the cookies in our clients array
    int nClients = m_Clients.GetSize();
    for (int i=0; i<nClients && !bFound; i++)
    {
      CComGITPtr<TSinkInterface>* pClient = m_Clients[i];
      if (pClient->GetCookie() == dwCookie)
      {
        bFound = TRUE;
        hr = pClient->Revoke();
        m_Clients.RemoveAt(i);
        delete pClient;
      }
    }

	  return SUCCEEDED(hr) ? S_OK : CONNECT_E_NOCONNECTION;
  }

	STDMETHOD(EnumConnections)(IEnumConnections** ppEnum)
  {
    //Validate our parameters
	  if (ppEnum == NULL)
		  return E_POINTER;

    //Set the output parameter to a default value
	  *ppEnum = NULL;

    //Create the COM object which will handle the COM interface being returned
	  CComObject<CComEnumConnections>* pEnum = NULL;
	  ATLTRY(pEnum = new CComObject<CComEnumConnections>)
	  if (pEnum == NULL)
		  return E_OUTOFMEMORY;

    //Serialize access to the clients array
	  CComCritSecLock<CComAutoCriticalSection> lock(m_CPMTCritSec);
	  
	  //Allocate the array
	  int nClients = m_Clients.GetSize();
	  CONNECTDATA* pcd = NULL;
	  ATLTRY(pcd = new CONNECTDATA[nClients])
	  if (pcd == NULL)
	  {
		  delete pEnum;
		  return E_OUTOFMEMORY;
	  }

	  //Copy the valid CONNECTDATA's
	  CONNECTDATA* pend = pcd;
	  HRESULT hr = S_OK;
	  for (int i=0; i<nClients; i++)
	  {
		  CComGITPtr<TSinkInterface>* pClient = m_Clients[i];
		  CComPtr<TSinkInterface> sinkInterface;
	    hr = pClient->CopyTo(&sinkInterface);
	    if (SUCCEEDED(hr))
	    {
	      hr = sinkInterface.QueryInterface(&pend->pUnk);
	      if (SUCCEEDED(hr))
	      {
	        //Also set the cookie value and move on to the next element in the output CONNECTDATA array
		      pend->dwCookie = pClient->GetCookie();
		      pend++;
		    }
	    }
	  }
	  //don't copy the data, but transfer ownership to it
	  hr = pEnum->Init(pcd, pend, NULL, AtlFlagTakeOwnership);
	  lock.Unlock();
	  if (FAILED(hr))
	    delete pEnum;
	  else
	  {
	    hr = pEnum->_InternalQueryInterface(IID_IEnumConnections, reinterpret_cast<void**>(ppEnum));
	    if (FAILED(hr))
		    delete pEnum;
		}
	  return hr;
  }

	IDispatch* GetInterfaceAt(int nConnectionIndex)
  {
    //It is assumed that client code has already taken a lock on m_CPMTCritSec
    //at this point to ensure we can safely access m_Clients directly

	  CComPtr<TSinkInterface> sinkInterface;
	  HRESULT hr = m_Clients[nConnectionIndex]->CopyTo(&sinkInterface);
	  ATLASSERT(SUCCEEDED(hr));
	  
	  //What will be the return value from this function (assume the worst)
	  IDispatch* pDisp = NULL;
	  
	  hr = sinkInterface.QueryInterface(&pDisp);
	  ATLASSERT(SUCCEEDED(hr));

	  return pDisp;
  }

//Member variables
	CSimpleArray<CComGITPtr<TSinkInterface>* >  m_Clients;     //The array of GIT interface pointers we are managing. Note we are 
	                                                           //storing heap pointers on the array to give us explicit control
	                                                           //over when and how copy constructors and operator= methods are 
	                                                           //being called as we add a element to the array
	CComGlobalsThreadModel::AutoCriticalSection m_CPMTCritSec; //The critical section to serialize access to "m_Clients" above
};

#endif //__CPIMPLMT_H__
