//////////////////// Includes ///////////////////////////////////

#include "stdafx.h"
#include "GPSCom.h"
#include "GPSConnection.h"
#include "NMEA.h"


///////////////////// Macros / Defines //////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


//////////////////// Implementation /////////////////////////////

CGPSConnection::CGPSConnection() : m_pMonitorThread(NULL),
                                   m_nDataBits(8),
                                   m_nBaudRate(4800),
                                   m_StopBits(GPS_1_STOP_BIT),
                                   m_Parity(GPS_PARITY_NONE),
                                   m_nPort(0),
                                   m_pRMCSentence(NULL),
                                   m_lWantThreadShutdown(FALSE),
                                   m_pGGASentence(NULL)
{
}

CGPSConnection::~CGPSConnection()
{
  //Ensure we close down everything before we are finished
  Close();

  //Synchronize access to the connections array
  CMyApp* pApp = GetApp();
  CComCritSecLock<CComAutoCriticalSection> lock(pApp->m_csConnections);

  //Remove this instance from the app list
  BOOL bFound = FALSE;
  int nConnections = pApp->m_Connections.GetSize();
  for (int i=0; i<nConnections && !bFound; i++)
  {
    CComObject<CGPSConnection>* pConnection = static_cast<CComObject<CGPSConnection>*>(pApp->m_Connections[i]);
    ASSERT(pConnection);
    if (pConnection->m_nPort == m_nPort)
    {
      pApp->m_Connections.RemoveAt(i);
      bFound = TRUE;
    }
  }
}

HRESULT CGPSConnection::FinalConstruct()
{
  //Create the internal RMC sentence member variable
  ATLASSERT(m_pRMCSentence == NULL);
  HRESULT hr = CComObject<CRMCSentence>::CreateInstance(&m_pRMCSentence);
  if (SUCCEEDED(hr))
  {
    //to keep the object alive
    m_pRMCSentence->AddRef();
  }

  if (SUCCEEDED(hr))
  {
    //Create the internal GGA sentence member variable
    ATLASSERT(m_pGGASentence == NULL);
    hr = CComObject<CGGASentence>::CreateInstance(&m_pGGASentence);
    if (SUCCEEDED(hr))
    {
      //to keep the object alive
      m_pGGASentence->AddRef();
    }
  }

  return hr;
}

void CGPSConnection::FinalRelease()
{
  //Release our two internal interfaces
  if (m_pRMCSentence)
  {
	  m_pRMCSentence->Release();
    m_pRMCSentence = NULL;
  }
  if (m_pGGASentence)
  {
	  m_pGGASentence->Release();
    m_pGGASentence = NULL;
  }
}

HRESULT CGPSConnection::Open(long nPort, long nDataBits, long nBaudRate, GPS_STOP_BITS StopBits, GPS_PARITY Parity, const CComBSTR& bstrInitString)
{
  //Open up the serial port
  try
  {
    //Validate out state
    ATLASSERT(!m_Port.IsOpen());

    CSerialPort::Parity LocalParity(CSerialPort::NoParity);
    switch (Parity)
    {
      case GPS_PARITY_NONE:  
      {
        LocalParity = CSerialPort::NoParity;
        break;
      }
      case GPS_PARITY_ODD:   
      {
        LocalParity = CSerialPort::OddParity;
        break;
      }
      case GPS_PARITY_EVEN:  
      {
        LocalParity = CSerialPort::EvenParity;
        break;
      }
      case GPS_PARITY_MARK:  
      {
        LocalParity = CSerialPort::MarkParity;
        break;
      }
      case GPS_PARITY_SPACE: 
      {
        LocalParity = CSerialPort::SpaceParity;
        break;
      }
      default:               
      {
        ATLASSERT(FALSE);                       
        break;
      }
    }    

    CSerialPort::StopBits LocalStopBits(CSerialPort::OneStopBit);
    switch (Parity)
    {
      case GPS_1_STOP_BIT:
      {
        LocalStopBits = CSerialPort::OneStopBit; 
        break;
      }
      case GPS_1POINT5_STOP_BITS: 
      {
        LocalStopBits = CSerialPort::OnePointFiveStopBits;
        break;
      }
      case GPS_2_STOP_BITS:
      {
        LocalStopBits = CSerialPort::TwoStopBits;
        break;
      }
      default:  
      {
        ATLASSERT(FALSE);
        break;
      }
    }    
    m_Port.Open(nPort, nBaudRate, LocalParity, (BYTE) nDataBits, LocalStopBits);

    //Setup return immediately from any ReadFile calls
    m_Port.Set0ReadTimeout();
  }
  catch(CSerialException* pEx)
  {
    //Form the rich error information
    Error(IDS_FAIL_OPEN_PORT, IID_IGPSPort);

    HRESULT hr = MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, pEx->m_dwError);
    pEx->Delete();
    m_Port.Close();
    return hr;
  }

  //Create the monitoring thread
  ATLASSERT(m_pMonitorThread == NULL);
  m_pMonitorThread = AfxBeginThread(_MonitorThread, this, THREAD_PRIORITY_IDLE, 0, CREATE_SUSPENDED);  
  if (m_pMonitorThread == NULL)
  {
    //Form the rich error information
    Error(IDS_FAIL_CREATE_THREAD, IID_IGPSPort);

    Close();
    return MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, GetLastError());
  }

  //We are looking after the shutdown of the thread
  m_pMonitorThread->m_bAutoDelete = FALSE;
  m_pMonitorThread->ResumeThread();

  //Store away the member variables
  m_nDataBits = nDataBits;
  m_nBaudRate = nBaudRate;
  m_StopBits = StopBits;
  m_Parity = Parity;
  m_nPort = nPort;
  m_bstrInitString = bstrInitString;

  //everything ok, return success indicator
  return S_OK;
}

void CGPSConnection::Close()
{
  if (m_pMonitorThread)
  {
    //Request the worker thread to exit
    InterlockedExchange(&m_lWantThreadShutdown, TRUE);

    //Synchronise with the exit of the monitor thread
    WaitForSingleObject(m_pMonitorThread->m_hThread, INFINITE);
    delete m_pMonitorThread;
    m_pMonitorThread = NULL;
  }

  //Close down the comms port
  m_Port.Close();
}

UINT CGPSConnection::_MonitorThread(void* pVoid)
{
  //Validate our parameters
  ATLASSERT(pVoid);

  //Initialize COM in this thread as we will be using connection points from it
  CoInitialize(NULL);

  //Convert from the SDK world to the C++ world
  CGPSConnection* pThis = static_cast<CGPSConnection*>(pVoid);
  ATLASSERT(pThis);
  UINT nReturn = pThis->MonitorThread();

  //Unitialize COM in this thread
  CoUninitialize();

  return nReturn;
}

UINT CGPSConnection::MonitorThread()
{
  //Send the init string if we have one
  if (m_bstrInitString.Length())
  {
    CW2A szInit(m_bstrInitString.operator BSTR());
    DWORD dwInitLength = static_cast<DWORD>(strlen(szInit));
    try
    {
      m_Port.Write(szInit.operator LPSTR(), dwInitLength); 
    }
    catch(CSerialException* pEx)
    {
      pEx->Delete();
    }
  }

  //Sit in a tight loop reading data from the port
  CString sSerialString;
  char szBuffer[100];
  while (!m_lWantThreadShutdown)
  {
    //Read whatever data is present from the serial port
    DWORD nBytesRead = 0;
    try
    {
      nBytesRead = m_Port.Read(szBuffer, sizeof(szBuffer) - 1);
    }
    catch(CSerialException* pEx)
    {
      pEx->Delete();
    }
    
    //If any data was read from the serial port, then parse the received data
    if (nBytesRead > 0) 
    {
      //Null terminate the received data
      szBuffer[nBytesRead] = '\0';

      //Append the newly read data to the buffer
      sSerialString += szBuffer;

      //See if we've got a sentence yet
      int nLocationLineFeed = sSerialString.Find(_T('\n'));
      if (nLocationLineFeed == -1) //Try a carriage return
        nLocationLineFeed = sSerialString.Find(_T('\r'));
      while (nLocationLineFeed != -1) 
      {
        //Yes, there is a line feed in the string
        CString sSentence(sSerialString.Left(nLocationLineFeed + 1));

        //Now trim off the sentence from what came in from the serial port
        sSerialString = sSerialString.Right(sSerialString.GetLength() - (nLocationLineFeed+1));
        nLocationLineFeed = sSerialString.Find(_T('\n'));
        if (nLocationLineFeed == -1) //Try a carriage return
          nLocationLineFeed = sSerialString.Find(_T('\r'));

        //determine which parser to call
        if (sSentence.Left(6).Compare(_T("$GPRMC")) == 0) 
        {
          CNmeaSentence sentence(sSentence);
          CRMCResponse response;
          if (response.Parse(sSentence))
          {
            ATLASSERT(m_pRMCSentence);
            m_pRMCSentence->m_Valid = (response.m_DataValid == NMEA_True);
            m_pRMCSentence->m_nLongitude = CNmeaSentence::NmeaAngleToHundrethsOfSeconds(response.m_Longitude.Value);
            if (response.m_Longitude.Easting == NMEA_East)
              m_pRMCSentence->m_bEasting = VARIANT_TRUE;
            else
              m_pRMCSentence->m_bEasting =  VARIANT_FALSE;
            m_pRMCSentence->m_nLatitude = CNmeaSentence::NmeaAngleToHundrethsOfSeconds(response.m_Latitude.Value);
            if (response.m_Latitude.Northing == NMEA_North)
              m_pRMCSentence->m_bNorthing = VARIANT_TRUE;
            else
              m_pRMCSentence->m_bNorthing = VARIANT_FALSE;
            m_pRMCSentence->m_dblBearing = response.m_Bearing;
            m_pRMCSentence->m_dblSpeed = response.m_SpeedOverGroundKnots;
            m_pRMCSentence->m_Date = COleDateTime(response.m_Date.wYear, response.m_Date.wMonth, response.m_Date.wDay,
                                                  response.m_Time.wHour, response.m_Time.wMinute, response.m_Time.wSecond);

            //Fire the event
            Fire_OnRMCSentence();
          }
        }
        else if (sSentence.Left(6).Compare(_T("$GPGGA")) == 0) 
        {
          CNmeaSentence sentence(sSentence);
          CGGAResponse response;
          if (response.Parse(sSentence))
          {
            ATLASSERT(m_pGGASentence);
            m_pGGASentence->m_nSatellites = response.m_nSatellites;
            m_pGGASentence->m_dblAltitude = response.m_AntennaAltitudeMeters;
            switch (response.m_Fix)
            {
              case NMEA_FIX_NOT_AVAILABLE:        
              {
                m_pGGASentence->m_Fix = GPS_QUALITY_FIX_NOT_AVAILABLE; 
                break;
              }
              case NMEA_QUALITY_FIX:              
              {
                m_pGGASentence->m_Fix = GPS_QUALITY_FIX;               
                break;
              }
              case NMEA_QUALITY_DIFFERENTIAL_FIX: 
              {
                m_pGGASentence->m_Fix = GPS_QUALITY_DIFFERENTIAL_FIX;  
                break;
              }
              default: 
              {
                ATLASSERT(FALSE);          
                break;
              }
            }

            //Fire the event
            Fire_OnGGASentence();
          }
        }
        else
        {
          //Fire the event for a NMEA sentence which we have not parsed
          BSTR bstrSentence = sSentence.AllocSysString();
          Fire_OnUnknownSentence(bstrSentence);
          SysFreeString(bstrSentence);
        }
      }
    }

    //Sleep for a time before we loop around, this avoids heavy CPU utilization. A future
    //release may use Overlapped IO thus preventing the need for this Busy loop with Sleep calls
    Sleep(100);
  }

  return 0;
}

STDMETHODIMP CGPSConnection::get_Parity(GPS_PARITY* pVal)
{
  //validate our parameters
	if (pVal == NULL)
    return E_POINTER;

  *pVal = m_Parity;

	return S_OK;
}

STDMETHODIMP CGPSConnection::get_StopBits(GPS_STOP_BITS* pVal)
{
	//validate our parameters
  if (pVal == NULL)
    return E_POINTER;

  *pVal = m_StopBits;

	return S_OK;
}

STDMETHODIMP CGPSConnection::get_BaudRate(long* pVal)
{
  //validate our parameters
  if (pVal == NULL)
    return E_POINTER;

  *pVal = m_nBaudRate;

	return S_OK;
}

STDMETHODIMP CGPSConnection::get_DataBits(long* pVal)
{
  //validate our parameters
	if (pVal == NULL)
    return E_POINTER;

  *pVal = m_nDataBits;

	return S_OK;
}

STDMETHODIMP CGPSConnection::get_Port(long* pVal)
{
  //validate our parameters
	if (pVal == NULL)
    return E_POINTER;

  *pVal = m_nPort;

	return S_OK;
}

STDMETHODIMP CGPSConnection::get_RMC(IRMCSentence** ppVal)
{
  //validate our parameters
	if (ppVal == NULL)
    return E_POINTER;

  //Query for the specific early bound interface which we want to hand back
  ATLASSERT(m_pRMCSentence);
  return m_pRMCSentence->QueryInterface(IID_IRMCSentence, reinterpret_cast<void**>(ppVal));
}

STDMETHODIMP CGPSConnection::get_GGA(IGGASentence** ppVal)
{
  //validate our parameters
	if (ppVal == NULL)
    return E_POINTER;

  //Query for the specific early bound interface which we want to hand back
  ATLASSERT(m_pGGASentence);
  return m_pGGASentence->QueryInterface(IID_IGGASentence, reinterpret_cast<void**>(ppVal));
}

STDMETHODIMP CGPSConnection::get_InitString(BSTR* pVal)
{
  //Validate out parameters
  if (pVal == NULL)
    return E_POINTER;

	return m_bstrInitString.CopyTo(pVal);
}
